;; Generated package description from memory-usage.el  -*- no-byte-compile: t -*-
(define-package "memory-usage" "0.2.0.20201201.223908" "Analyze the memory usage of Emacs in various ways" 'nil :commit "c751101ba3adaca67a37977562907557963007bb" :url "https://elpa.gnu.org/packages/memory-usage.html" :authors '(("Stefan Monnier" . "monnier@iro.umontreal.ca")) :maintainer '("Stefan Monnier" . "monnier@iro.umontreal.ca") :keywords '("maint"))
